import os, torch, cv2, pandas as pd
from tqdm import tqdm
import numpy as np
from text_encoder import TextEncoder
from transformers import DistilBertTokenizer
from model_contrastive import DLOPerceiver
from metrics import get_metrics


CHECKPOINT_NAME = "unique-rain-123.pt"

PATH_TEST = "/home/lar/dev24"
FOLDER = ["TEST_C3_16_9"]

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
print("DEVICE: ", DEVICE)

COLORS = ["red", "green", "blue", "black", "white", "gray", "yellow", "orange", "purple", "striped"]
OBJECTS = ["cable", "rope"]
POSITIONS = ["top", "bottom"]


class Evaluator:

    def __init__(self, checkpoint_path, dataset_path) -> None:

        model_dict = torch.load(checkpoint_path)
        model_config = model_dict["config"]
        print("model_config: ", model_config)

        self.img_h = model_config["img_h"]
        self.img_w = model_config["img_w"]

        self.model = DLOPerceiver(
            iterations=model_config["iterations"],
            n_latents=model_config["n_latents"],
            latent_dim=model_config["latent_dim"],
            depth=model_config["depth"],
            dropout=model_config["dropout"],
            img_encoder_type=model_config["img_encoder_type"],
            noise_aug=False,
        )

        self.model.load_state_dict(model_dict["model"])
        self.model.to(device=DEVICE)
        self.model.eval()

        #################
        self.rgb_folder = "test_imgs"
        self.masks_folder = "test_labels"
        self.text_folder = "test_text"

        self.rgb_path = os.path.join(dataset_path, self.rgb_folder)
        self.masks_path = os.path.join(dataset_path, self.masks_folder)
        self.text_path = os.path.join(dataset_path, self.text_folder)
        self.data = sorted(os.listdir(self.rgb_path))

        self.tokenizer = DistilBertTokenizer.from_pretrained("distilbert-base-uncased")
        self.text_encoder = TextEncoder(model_name="distilbert-base-uncased")

        self.dlo_groups = ["only_cable", "only_rope", "mixed"]
        self.scene_complexity = ["simple", "difficult"]

    def pre_process(self, img):
        if len(img.shape) == 2:
            img = np.expand_dims(img, axis=2)
        img = img.transpose((2, 0, 1))
        if img.max() > 1:
            img = img / 255
        return img

    def get_sample(self, file_name):

        ###############################
        if "C" in file_name:
            group = "only_cable"
        elif "R" in file_name:
            group = "only_rope"
        else:
            group = "mixed"

        n_file_name = int(file_name.split("_")[-1].split(".")[0])
        if n_file_name == 1:
            scene_complexity = "simple"
        else:
            scene_complexity = "difficult"

        # rgb
        img = cv2.imread(os.path.join(self.rgb_path, file_name), cv2.IMREAD_COLOR)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = cv2.resize(img, (self.img_w, self.img_h))
        img_t = torch.from_numpy(self.pre_process(img)).type(torch.FloatTensor)

        # mask
        mask = cv2.imread(os.path.join(self.masks_path, file_name), cv2.IMREAD_GRAYSCALE)
        mask = cv2.resize(mask, (self.img_w, self.img_h), interpolation=cv2.INTER_NEAREST)

        ##################
        # text
        text_path = os.path.join(self.text_path, file_name.replace(".png", ".csv"))
        text_df = pd.read_csv(text_path)

        return img_t, mask, text_df, (group, scene_complexity)

    def eval(self):

        combinations = [
            "tutto_ok",  # 0
            "wrong_obj",  # 1
            "wrong_color",  # 2
            "wrong_top_down",  # 3
            "wrong_obj_color",  # 4
            "wrong_obj_top_down",  # 5
            "wrong_color_top_down",  # 6
            "wrong_all",  # 7
        ]
        columns = ["id", "category_dlo", "categoy_scene", "prompt_type", "obj", "color", "pos", "iou", "dice", "conf"]
        values = []
        for batch_i, batch_path in enumerate(tqdm(self.data)):

            img_t, mask, text_df, category = self.get_sample(batch_path)
            category_dlo = category[0]
            category_scene = category[1]

            img = img_t.unsqueeze(0).to(device=DEVICE)

            ###############################

            scene_colors = []
            for row in text_df.iterrows():
                row = row[1]
                color = row["real_color"] if not pd.isnull(row["real_color"]) else None
                if color:
                    scene_colors.append(color)

            for row in text_df.iterrows():
                row = row[1]
                gray_value = row["gray_value"]
                label = np.zeros(mask.shape)
                label[mask == gray_value] = 1

                obj = row["object"] if not pd.isnull(row["object"]) else None
                color = row["real_color"] if not pd.isnull(row["real_color"]) else None
                top_down = row["top/down"] if not pd.isnull(row["top/down"]) else None

                obj_wrong = OBJECTS[0] if obj == OBJECTS[1] else OBJECTS[1]
                top_down_wrong = POSITIONS[0] if top_down == POSITIONS[1] else POSITIONS[1]

                colors_wrong = [c for c in COLORS if c not in scene_colors]

                for it, prompt_combination in enumerate(combinations):

                    if it == 0:  # tutto_ok
                        prompt = f"{obj} {color} {top_down}"
                    elif it == 1:  # wrong_obj
                        prompt = f"{obj_wrong} {color} {top_down}"
                    elif it == 2:  # wrong_color
                        prompt = f"{obj} {np.random.choice(colors_wrong)} {top_down}"
                    elif it == 3:  # wrong_top_down
                        prompt = f"{obj} {color} {top_down_wrong}"
                    elif it == 4:  # wrong_obj_color
                        prompt = f"{obj_wrong} {np.random.choice(colors_wrong)} {top_down}"
                    elif it == 5:  # wrong_obj_top_down
                        prompt = f"{obj_wrong} {color} {top_down_wrong}"
                    elif it == 6:  # wrong_color_top_down
                        prompt = f"{obj} {np.random.choice(colors_wrong)} {top_down_wrong}"
                    elif it == 7:  # wrong_all
                        prompt = f"{obj_wrong} {np.random.choice(colors_wrong)} {top_down_wrong}"

                    text_emb = self.tokenizer(prompt, return_tensors="pt")["input_ids"]
                    text_emb = self.text_encoder(text_emb).to(device=DEVICE)

                    ###############################
                    # pred
                    with torch.no_grad():
                        pred, conf = self.model(img, text_emb.unsqueeze(0))
                    pred = pred.sigmoid()
                    conf = conf.item()

                    pred = pred.squeeze().detach().cpu().numpy()
                    pred_mask = pred > 0.5

                    ###############################
                    # metrics
                    pred_mask = cv2.resize(pred_mask.astype(np.uint8), (label.shape[1], label.shape[0]))
                    score = get_metrics(pred_mask, label)

                    values.append(
                        [
                            batch_i,
                            category_dlo,
                            category_scene,
                            prompt_combination,
                            obj,
                            color,
                            top_down,
                            score["iou"],
                            score["dice"],
                            conf,
                        ]
                    )

        df = pd.DataFrame(values, columns=columns)
        return df


if __name__ == "__main__":
    project_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    print("project_path: ", project_path)
    checkpoint_path = os.path.join(project_path, "checkpoints", CHECKPOINT_NAME)

    for test_name in FOLDER:
        print("TEST: ", test_name)
        evalx = Evaluator(checkpoint_path, os.path.join(PATH_TEST, test_name))
        data_frame = evalx.eval()

        print(data_frame)

        data_frame.to_csv(f"{test_name}_wrong.csv", index=False)
