import bpy
import numpy as np


def generate_curve(id, points, radius=0.003, color=(0.7, 0.0, 0.0, 1.0), subdivide=True, rope=False, striped=False):

    cu = bpy.data.curves.new("curve_" + str(id), "CURVE")
    ob = bpy.data.objects.new("object_" + str(id), cu)
    scn = bpy.context.scene
    scn.collection.objects.link(ob)

    cu.dimensions = "3D"
    cu.resolution_u = 36
    cu.resolution_v = 36
    cu.bevel_resolution = 36
    cu.fill_mode = "FULL"
    cu.bevel_depth = radius

    # MAKE SPLINE
    spline = cu.splines.new("BEZIER")

    points = make_4D(points)

    spline.bezier_points.add(len(points) - 1)
    for n, pt in enumerate(points):
        spline.bezier_points[n].co = [pt[0], pt[1], pt[2]]
        spline.bezier_points[n].handle_right_type = "AUTO"
        spline.bezier_points[n].handle_left_type = "AUTO"

    bpy.context.view_layer.objects.active = ob
    so = bpy.context.active_object

    # create the material
    new_mat = bpy.data.materials.new(name="material_" + str(id))
    so.data.materials.append(new_mat)

    new_mat.use_nodes = True

    # get all material nodes
    nodes = new_mat.node_tree.nodes
    links = new_mat.node_tree.links

    material_output = nodes.get("Material Output")
    bsdf = nodes.get("Principled BSDF")
    links.new(bsdf.outputs[0], material_output.inputs[0])

    bsdf.inputs[0].default_value = color

    # saturation (only for rope)
    if rope:
        bsdf.inputs[1].default_value = np.random.uniform(0.8, 1.0)

    # Specular
    bsdf.inputs[5].default_value = np.random.uniform(0.0, 0.5)

    # Roughness
    bsdf.inputs[7].default_value = np.random.uniform(0.2, 0.8)

    # IOR
    bsdf.inputs[14].default_value = np.random.uniform(1.3, 1.7)

    if not rope and striped:
        mapping = nodes.new("ShaderNodeMapping")
        mapping.inputs[3].default_value = (1.0, 1.5, 1.0)
        mapping.vector_type = "TEXTURE"

        tex_coord = nodes.new("ShaderNodeTexCoord")

        wave_tex = nodes.new("ShaderNodeTexWave")
        wave_tex.inputs[1].default_value = 1.0
        wave_tex.inputs[2].default_value = 0.0
        wave_tex.inputs[3].default_value = 0.0
        wave_tex.inputs[4].default_value = 0.0
        wave_tex.inputs[5].default_value = 0.0
        wave_tex.inputs[6].default_value = 0.0
        wave_tex.wave_type = "BANDS"
        wave_tex.wave_profile = "SAW"
        wave_tex.bands_direction = "Y"

        checker_tex = nodes.new("ShaderNodeTexChecker")
        checker_tex.inputs[1].default_value = color  # green
        checker_tex.inputs[2].default_value = (0.6, 0.5, 0.0, 1.0)  # yellow
        checker_tex.inputs[3].default_value = 2.0

        link1 = links.new(tex_coord.outputs[2], mapping.inputs[0])
        link2 = links.new(mapping.outputs[0], wave_tex.inputs[0])
        link3 = links.new(wave_tex.outputs[0], checker_tex.inputs[0])
        link4 = links.new(checker_tex.outputs[0], bsdf.inputs[0])

    if subdivide:
        bpy.ops.object.mode_set(mode="EDIT")
        bpy.ops.curve.select_all(action="SELECT")
        bpy.ops.curve.subdivide(number_cuts=20)
        bpy.ops.object.mode_set(mode="OBJECT")

    return ob


def make_4D(points):
    if points.shape[1] == 2:
        return [(p[0], p[1], 0.0, 1) for p in points]
    else:
        return [(p[0], p[1], p[2], 1) for p in points]


def generate_rope_from_curve(curve_obj_name, curve_len, radius, color=(0.7, 0.0, 0.0, 1.0), num_chords=2):

    curve_len_adjusted = curve_len * 1.1

    # Adjust curve length with a random factor
    aug_factor = np.random.uniform(1.0, 1.5)
    curve_len_adjusted *= aug_factor
    iters = round(100 / aug_factor)

    # Create circles along the curve
    bpy.ops.mesh.primitive_circle_add(location=(0, 0, 0))
    bpy.ops.transform.resize(value=(radius, radius, radius))
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.transform.rotate(value=np.pi / 2, orient_axis="X")
    bpy.ops.transform.translate(value=(radius, 0, 0))
    bpy.ops.object.mode_set(mode="OBJECT")
    for i in range(1, num_chords):
        bpy.ops.object.duplicate_move(OBJECT_OT_duplicate=None, TRANSFORM_OT_translate=None)
        ob = bpy.context.active_object
        ob.rotation_euler = (0, 0, i * (2 * np.pi / num_chords))

    # Create circles along the curve
    bpy.ops.mesh.primitive_circle_add(location=(0, 0, 0))
    bpy.ops.transform.resize(value=(radius, radius, radius))
    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.transform.rotate(value=np.pi / 2, orient_axis="X")
    bpy.ops.transform.translate(value=(radius, 0, 0))
    bpy.ops.object.mode_set(mode="OBJECT")
    for i in range(1, num_chords):
        bpy.ops.object.duplicate_move(OBJECT_OT_duplicate=None, TRANSFORM_OT_translate=None)
        ob = bpy.context.active_object
        ob.rotation_euler = (0, 0, i * (2 * np.pi / num_chords))

    # Join circles into a single object
    bpy.ops.object.select_all(action="DESELECT")
    for o in bpy.data.objects:
        if "Circle" in o.name:
            o.select_set(True)
    bpy.ops.object.join()

    # Apply screw modifier to create the rope
    bpy.ops.object.modifier_add(type="SCREW")
    rope = bpy.context.object
    rope.scale = (radius, radius, 0.01)
    rope.rotation_euler = (0, np.pi / 2, 0)
    rope.modifiers["Screw"].screw_offset = curve_len_adjusted
    rope.modifiers["Screw"].iterations = iters
    rope.modifiers["Screw"].use_smooth_shade = True
    rope.modifiers["Screw"].steps = 36

    # Apply curve modifier to shape the rope along the curve
    bpy.ops.object.modifier_add(type="CURVE")
    rope.modifiers["Curve"].object = bpy.data.objects[curve_obj_name]

    # Create and assign material to the rope
    new_mat = bpy.data.materials.new(name="material_rope")
    rope.data.materials.append(new_mat)
    new_mat.use_nodes = True

    # Configure material nodes
    material_output = new_mat.node_tree.nodes.get("Material Output")
    bsdf = new_mat.node_tree.nodes.get("Principled BSDF")
    new_mat.node_tree.links.new(bsdf.outputs[0], material_output.inputs[0])
    bsdf.inputs[0].default_value = color

    # change name to circle
    rope.name = curve_obj_name + "_rope"
