import blenderproc as bproc
import numpy as np
import pickle, os, sys, copy, cv2, random
import bpy

# import subprocess
# python_exe = "/home/alessio/blender/blender-3.5.1-linux-x64/3.5/python/bin/python3.10"
# subprocess.call([python_exe, "-m", "ensurepip"])
# subprocess.call([python_exe, "-m", "pip", "install", "--upgrade", "pip"])
# subprocess.call([python_exe, "-m", "pip", "install", "shapely"])
# subprocess.call([python_exe, "-m", "pip", "install", "albumentations"])

import albumentations as aug

script_path = os.path.dirname(os.path.realpath(__file__))
sys.path.append(script_path)

from rendering_scene_gen import SceneGen, CameraSampler

CONFIG = {
    "output_path": "new_renderings",
    "z_threshold": 0.02,
    "make_rope1": random.choice([True, False]),
    "make_rope2": random.choice([True, False]),
    "radius": [0.003, 0.008],
    "haven_texture_enabled": False,
    "haven_texture_path": "path_to_haven_textures",
    "haven_world_hdri_enabled": False,
    "haven_hdri_path": "path_to_hdri",
    "abstract_texture_enabled": True,
    "abstract_texture_path": "path_to_abstract_backgrounds",
    "same_color": False,
}

print("CONFIG: ", CONFIG)


COLORS = {
    "red": (0.7, 0.0, 0.0, 1.0),
    "green": (0.0, 0.2, 0.0, 1.0),
    "blue": (0.0, 0.0, 0.7, 1.0),
    "black": (0.0, 0.0, 0.0, 1.0),
    "white": (0.9, 0.9, 0.9, 1.0),
    "gray": (0.3, 0.3, 0.3, 1.0),
    "yellow": (0.6, 0.5, 0.0, 1.0),
    "orange": (0.95, 0.2, 0.0, 1.0),
    "purple": (0.5, 0.0, 0.5, 1.0),
    "striped": (0.0, 0.2, 0.0, 1.0),  # green with yellow
}


class Camera320:
    H = 180
    W = 320
    K = [150, 0.0, W / 2, 0.0, 150, H / 2, 0.0, 0.0, 1.0]


class Camera640:
    H = 360
    W = 640
    K = [300, 0.0, W / 2, 0.0, 300, H / 2, 0.0, 0.0, 1.0]


class Camera1280:
    H = 720
    W = 1280
    K = [600, 0.0, W / 2, 0.0, 600, H / 2, 0.0, 0.0, 1.0]


TRANSFORM = aug.Compose(
    [
        aug.Flip(p=0.5),
        aug.FancyPCA(p=0.8),
        aug.GaussNoise(p=0.5),
        aug.HueSaturationValue(p=0.8),
        aug.RGBShift(p=0.5),
        aug.RandomBrightnessContrast(p=0.8),
    ],
    p=1,
)


class Rendering:
    def __init__(self):
        self.camera_sampler = CameraSampler(Camera640)
        self.gen = SceneGen(config=CONFIG, colors=COLORS, same_color=CONFIG["same_color"])

        self.output_path = CONFIG["output_path"]
        os.makedirs(self.output_path, exist_ok=True)

    def augment_background_image(self, image_path):
        # convert image to np array
        img = cv2.imread(image_path, cv2.IMREAD_COLOR)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

        # agument image
        img_aug = TRANSFORM(image=img)["image"]
        img_aug = cv2.cvtColor(img_aug, cv2.COLOR_RGB2RGBA) / 255.0

        # create output image
        output_image = bpy.data.images.new("background.png", width=img.shape[1], height=img.shape[0])
        output_image.file_format = "JPEG"
        output_image.filepath = "background.png"

        # convert np array back to image
        output_image.pixels = img_aug.ravel()
        return output_image

    def exec(self):
        bproc.init()
        bpy.context.scene.world["category_id"] = 0

        if CONFIG["haven_world_hdri_enabled"]:
            hdri_path = bproc.loader.get_random_world_background_hdr_img_path_from_haven(CONFIG["haven_hdri_path"])
            bproc.world.set_world_background_hdr_img(hdri_path)

        else:
            # sun
            sun = bproc.types.Light(light_type="SUN")
            random_energy = np.random.uniform(0.5, 3.0)
            sun.set_energy(random_energy)

        #######################################################
        # generate scene
        gt_data, intersections_poi = self.gen.main2()

        #######################################################
        # camera
        bproc.camera.set_intrinsics_from_K_matrix(
            self.camera_sampler.camera_matrix, self.camera_sampler.camera_width, self.camera_sampler.camera_height
        )
        poi = intersections_poi[0]
        poi = [poi[0], poi[1], 0.0]
        cam2world_matrix, T = self.camera_sampler.sample(poi)
        bproc.camera.add_camera_pose(cam2world_matrix)

        #######################################################
        # supporting plane
        plane = bproc.object.create_primitive("PLANE", scale=[1.0, 1.0, 1], location=poi)
        plane.set_name("ground_plane0")
        plane.set_cp("category_id", 0)

        if CONFIG["haven_texture_enabled"]:
            print("Using haven texture")
            texture = bproc.loader.load_haven_mat(folder_path=CONFIG["haven_texture_path"], return_random_element=True)
        elif CONFIG["abstract_texture_enabled"]:
            print("Using abstract texture")
            files = os.listdir(CONFIG["abstract_texture_path"])
            random_file_path = os.path.join(CONFIG["abstract_texture_path"], np.random.choice(files))
            img_plane = self.augment_background_image(random_file_path)
            texture = bproc.material.create_material_from_texture(img_plane, "plane_mat")
        else:
            print("No texture provided")
        plane.replace_materials(texture)

        #######################################################
        # Render the scene
        data = bproc.renderer.render()
        data.update(bproc.renderer.render_segmap(map_by=["class", "instance", "name"]))

        #######################################################
        # save
        self.save_on_disk(data, gt_data, T)

    def save_on_disk(self, data, gt_data, camera_pose):
        colors = np.array(data["colors"]).squeeze()
        instance_segmaps = np.array(data["instance_segmaps"]).squeeze()

        # fix instance segmaps
        unique_ids = np.unique(instance_segmaps)
        if CONFIG["make_rope1"] and not CONFIG["make_rope2"]:
            instance_segmaps[instance_segmaps == 3] = 1
        elif CONFIG["make_rope2"] and not CONFIG["make_rope1"]:
            instance_segmaps[instance_segmaps == 3] = 2
        elif CONFIG["make_rope1"] and CONFIG["make_rope2"]:
            instance_segmaps[instance_segmaps == 3] = 1
            instance_segmaps[instance_segmaps == 4] = 2
        instance_segmaps[instance_segmaps == unique_ids[-1]] = 0

        output = {}
        output["pose"] = camera_pose
        output["gt"] = gt_data
        output["camera_matrix"] = self.camera_sampler.camera_matrix

        # save incrementally
        new_id = self.get_new_pickle_id()
        folder_path = os.path.join(self.output_path, str(new_id))

        os.makedirs(folder_path, exist_ok=True)
        cv2.imwrite(os.path.join(folder_path, "color.png"), cv2.cvtColor(colors, cv2.COLOR_RGB2BGR))
        cv2.imwrite(os.path.join(folder_path, "instances.png"), instance_segmaps)

        with open(os.path.join(folder_path, "gt.pkl"), "wb") as handle:
            pickle.dump(output, handle, protocol=pickle.HIGHEST_PROTOCOL)

    def get_new_pickle_id(self):
        # list dirs
        existing_data = os.listdir(self.output_path)

        # check if dirs
        existing_data = [f for f in existing_data if os.path.isdir(os.path.join(self.output_path, f))]

        if existing_data:
            last_id = max([int(d) for d in existing_data])
        else:
            last_id = -1

        return last_id + 1


if __name__ == "__main__":
    r = Rendering()
    r.exec()
