import blenderproc as bproc

import bpy
import numpy as np
import shapely
from termcolor import cprint

import numpy as np
import cv2
import copy

from rendering_curve_rope import generate_curve, generate_rope_from_curve


class CameraSampler:
    def __init__(self, camera_obj):
        self.camera_matrix = np.array(camera_obj.K).reshape(3, 3)
        self.camera_height = camera_obj.H
        self.camera_width = camera_obj.W
        self.camera_dist = np.array([0.0, 0.0, 0.0, 0.0, 0.0], dtype=np.float32)

        self.borders_offset = self.camera_width / 10

        self.radius_min = 0.3
        self.radius_max = 0.5
        self.elevation_min = 60
        self.elevation_max = 90

    def camera_pose(self, poi):
        random_x_offset = np.random.choice([+1, -1]) * np.random.uniform(0.1, 0.2)
        random_y_offset = np.random.choice([+1, -1]) * np.random.uniform(0.1, 0.2)
        random_z_offset = np.random.uniform(0.3, 0.5)
        location = np.array([poi[0] + random_x_offset, poi[1] + random_y_offset, random_z_offset])

        location_2 = location.copy()
        location_2[0] = location_2[0] + np.random.uniform(-0.1, 0.1)
        location_2[1] = location_2[1] + np.random.uniform(-0.1, 0.1)

        rotation_matrix = bproc.camera.rotation_from_forward_vec(
            poi - location_2, inplane_rot=np.random.uniform(-1.5709, 1.5709)
        )
        cam2world_matrix = bproc.math.build_transformation_mat(location, rotation_matrix)
        T = bproc.math.change_source_coordinate_frame_of_transformation_matrix(cam2world_matrix, ["X", "-Y", "-Z"])
        return cam2world_matrix, T

    def camera_pose_sphere(self, poi):
        inplace_rot = np.random.uniform(-1.5709, 1.5709)
        location = bproc.sampler.shell(
            center=poi,
            radius_min=self.radius_min,
            radius_max=self.radius_max,
            elevation_min=self.elevation_min,
            elevation_max=self.elevation_max,
            uniform_volume=False,
        )
        print("CAMERA LOCATION: ", location)
        rotation_matrix = bproc.camera.rotation_from_forward_vec(poi - location, inplane_rot=inplace_rot)
        cam2world_matrix = bproc.math.build_transformation_mat(location, rotation_matrix)
        T = bproc.math.change_source_coordinate_frame_of_transformation_matrix(cam2world_matrix, ["X", "-Y", "-Z"])
        return cam2world_matrix, T

    def project(self, camera_pose, points_3d):
        T = np.linalg.inv(copy.deepcopy(camera_pose))
        tvec = np.array(T[0:3, 3])
        rvec, _ = cv2.Rodrigues(T[:3, :3])

        point2d = cv2.projectPoints(np.array(points_3d), rvec, tvec, self.camera_matrix, self.camera_dist)
        point2d = point2d[0].squeeze()

        points_ref = []
        try:
            for p in point2d:
                i, j = [round(p[1]), round(p[0])]
                if i < self.camera_height and i >= 0 and j < self.camera_width and j >= 0:
                    points_ref.append([j, i])
        except:
            i, j = [round(point2d[1]), round(point2d[0])]
            if i < self.camera_height and i >= 0 and j < self.camera_width and j >= 0:
                points_ref.append([j, i])

        return points_ref

    def check_borders(self, point_img):
        return (
            point_img[0] > self.borders_offset
            and point_img[0] < (self.camera_width - self.borders_offset)
            and point_img[1] > self.borders_offset
            and point_img[1] < (self.camera_height - self.borders_offset)
        )

    def sample(self, poi, trials=5):
        for _ in range(trials):
            cam2world_matrix, T = self.camera_pose_sphere(poi)
            points_img = self.project(T, poi)

            flag_error = False
            for point_img in points_img:
                if not self.check_borders(point_img):
                    flag_error = True

            if not flag_error:
                return cam2world_matrix, T

        return None, None


class SceneGen:

    def __init__(self, config, colors, same_color=False):

        self.COLORS = colors
        self.radius_low, self.radius_high = config["radius"]
        self.z_threshold = config["z_threshold"]
        self.make_rope1 = config["make_rope1"]
        self.make_rope2 = config["make_rope2"]
        self.same_color = same_color

    def randomColor(self, exclude_key=None):
        k = list(self.COLORS.keys())

        if exclude_key:
            k.remove(exclude_key)

        x = np.random.choice(k)
        return x, self.COLORS[x]

    def randomRadius(self):
        values = np.arange(self.radius_low, self.radius_high + 0.001, step=0.001)
        return np.random.choice(values)

    def distance2D(self, point1, point2):
        return ((point1[0] - point2[0]) ** 2 + (point1[1] - point2[1]) ** 2) ** 0.5

    def closerListPointToList(self, point, list_points):
        min_id = [self.distance2D(point, p) for p in list_points]
        ind = np.argpartition(min_id, 2)[:2]
        return list_points[np.argmin(min_id)], list_points[ind[0]][:2], list_points[ind[1]][:2]

    def calcAngle(self, line0, line1):
        angle0 = np.arctan2(line0[0][1] - line0[1][1], line0[0][0] - line0[1][0])
        angle1 = np.arctan2(line1[0][1] - line1[1][1], line1[0][0] - line1[1][0])
        return np.fabs(angle0 - angle1)

    def get_random_path(self, num_rnd=20, step=0.05, zbounds=(0.01, 0.08)):
        step = np.random.uniform(low=0.075, high=0.15)

        init_x = np.random.uniform(low=0, high=0.2)
        init_y = np.random.uniform(low=0, high=0.2)
        init_z = np.random.uniform(low=0, high=0.05)

        path = np.array([(init_x, init_y, init_z)])

        while len(path) < num_rnd:
            p = np.random.uniform(low=0, high=1, size=(2,))
            z = np.random.uniform(low=-0.1, high=0.1)
            out = np.hstack(([p, z]))
            out = out / np.linalg.norm(out)
            new_point = path[-1] + step * out
            if new_point[2] < zbounds[1] and new_point[2] > zbounds[0]:
                path = np.append(path, new_point.reshape(1, 3), axis=0)
        return path

    def get_gt_dict(self, curves, colors, ids):
        gt_data = {}
        for it, curve in enumerate(curves):
            gt_data[curve.name] = {
                "points": [(p.co.x, p.co.y, p.co.z) for p in curve.splines[0].bezier_points],
                "color": colors[it],
                "id": ids[it],
            }
        return gt_data

    def rotate_shape_around_center(self, dlo_shape, angle):
        def get_rotation_matrix_from_angle(a):
            return np.array([[np.cos(a), -np.sin(a)], [np.sin(a), np.cos(a)]])

        rot_mat = get_rotation_matrix_from_angle(angle)

        dlo_shape_2d = dlo_shape[:, :2]
        center = np.mean(dlo_shape_2d, axis=0)
        dlo_shape[:, :2] = np.dot(rot_mat, (dlo_shape_2d - center).T).T + center
        return dlo_shape

    """
    def main(self):

        flag_intersect = False
        while not flag_intersect:
            good_intersect_points = []

            path_1 = self.get_random_path()
            path_2 = self.get_random_path()
            line1 = shapely.geometry.LineString([(p[0], p[1]) for p in path_1])
            line2 = shapely.geometry.LineString([(p[0], p[1]) for p in path_2])

            intersect_points = line1.intersection(line2)

            if bool(intersect_points):
                if type(intersect_points) == shapely.geometry.multipoint.MultiPoint:
                    points_out = [(point.x, point.y) for point in intersect_points.geoms]
                else:
                    points_out = [(intersect_points.x, intersect_points.y)]

                for _, p in enumerate(points_out):
                    path_1_point, p1_0, p1_1 = self.closerListPointToList(p, path_1)
                    path_2_point, p2_0, p2_1 = self.closerListPointToList(p, path_2)

                    arr1 = np.array([p1_0, p1_1])
                    arr2 = np.array([p2_0, p2_1])
                    angle = self.calcAngle(arr1, arr2)
                    angle = angle - np.pi if angle > np.pi else angle

                    z_diff = np.fabs(path_1_point[2] - path_2_point[2])
                    if (
                        z_diff > CONFIG["z_threshold"]
                        and angle > CONFIG["angle_threshold"]
                        and angle < (np.pi - CONFIG["angle_threshold"])
                    ):
                        flag_intersect = True
                        good_intersect_points.append(p)

        # make curve
        color1_name, color1_value = self.randomColor()
        color2_name, color2_value = self.randomColor()
        print("color1: ", color1_name, color1_value)
        print("color2: ", color2_name, color2_value)
        radius1 = self.randomRadius()
        radius2 = self.randomRadius()

        if CONFIG["make_rope1"]:
            radius1 = radius1 / 2
        if CONFIG["make_rope2"]:
            radius2 = radius2 / 2

        striped1 = False
        striped2 = False
        if color1_name == "striped":
            striped1 = True
        if color2_name == "striped":
            striped2 = True

        curve_obj1 = generate_curve(id=1, points=path_1, radius=radius1, color=color1_value, striped=striped1)
        curve_obj2 = generate_curve(id=2, points=path_2, radius=radius2, color=color2_value, striped=striped2)

        if CONFIG["make_rope1"]:
            length_curve1 = np.sum(np.linalg.norm(np.diff(path_1, axis=0), axis=1))
            num_chords_1 = 2
            radius1_rope = radius1 * 1.5
            obj1_name = curve_obj1.name
            generate_rope_from_curve(obj1_name, length_curve1, radius1_rope, color1_value, num_chords=num_chords_1)

        if CONFIG["make_rope2"]:
            length_curve2 = np.sum(np.linalg.norm(np.diff(path_2, axis=0), axis=1))
            num_chords_2 = 2  # np.random.choice([2, 4])
            radius2_rope = radius2 * 1.5
            obj2_name = curve_obj2.name
            generate_rope_from_curve(obj2_name, length_curve2, radius2_rope, color2_value, num_chords=num_chords_2)

        # save
        # bpy.ops.wm.save_as_mainfile(filepath="/home/lar/Downloads/scene.blend")

        # get GT points of each curve
        gt_data = self.get_gt_dict([curve_obj1.data, curve_obj2.data], [color1_name, color2_name], [1, 2])
        gt_data["curve_1"]["object_name"] = "cable" if not CONFIG["make_rope1"] else "rope"
        gt_data["curve_2"]["object_name"] = "cable" if not CONFIG["make_rope2"] else "rope"

        # Deselect all objects
        bpy.ops.object.select_all(action="DESELECT")

        # Convert to mesh
        for o in bpy.data.objects:
            o.select_set(True)
        bpy.ops.object.convert(keep_original=False)

        return gt_data, good_intersect_points
    """

    def main2(self):

        path_1 = self.get_random_path()
        path_2 = self.get_random_path()

        angle = np.random.choice([-1, 1]) * np.random.uniform(np.pi / 8, np.pi / 2)
        # print("angle: ", angle, np.degrees(angle))
        path_1_rot = self.rotate_shape_around_center(path_1, angle)

        line1 = shapely.geometry.LineString([(p[0], p[1]) for p in path_1_rot])
        line2 = shapely.geometry.LineString([(p[0], p[1]) for p in path_2])

        intersect_points = line1.intersection(line2)

        good_intersect_points = []
        z_threshold_gain = 2 if self.make_rope1 or self.make_rope2 else 1
        if bool(intersect_points):
            if type(intersect_points) == shapely.geometry.multipoint.MultiPoint:
                points_out = [(point.x, point.y) for point in intersect_points.geoms]
            else:
                points_out = [(intersect_points.x, intersect_points.y)]

            for _, p in enumerate(points_out):
                path_1_point, _, _ = self.closerListPointToList(p, path_1_rot)
                path_2_point, _, _ = self.closerListPointToList(p, path_2)

                z_diff = path_1_point[2] - path_2_point[2]
                if np.fabs(z_diff) > (self.z_threshold * z_threshold_gain):
                    pass
                elif z_diff < 0:
                    # path 1 is below path 2
                    path_2 = path_2 + np.array([0, 0, 0.02])
                else:
                    # path 2 is below path 1
                    path_1_rot = path_1_rot + np.array([0, 0, 0.02])

                good_intersect_points.append(p)

        # CURVE_1
        color1_name, color1_value = self.randomColor()
        radius1 = self.randomRadius()
        striped1 = True if color1_name == "striped" else False

        # CURVE_2
        if not self.same_color:
            color2_name, color2_value = self.randomColor()
        else:
            color2_name = color1_name
            color2_value = color1_value

        radius2 = self.randomRadius()
        striped2 = True if color2_name == "striped" else False

        cprint("OBJ1: {}, {}, {:.2f}, {}".format(color1_name, color1_value, radius1 * 1000, striped1), "green")
        cprint("OBJ2: {}, {}, {:.2f}, {}".format(color2_name, color2_value, radius2 * 1000, striped2), "green")

        # check if ropes are too small and disable them
        if radius1 < 0.004:
            self.make_rope1 = False
        if radius2 < 0.004:
            self.make_rope2 = False

        # reduce radius if rope to make it more realistic
        radius1 = radius1 / 2 if self.make_rope1 else radius1
        radius2 = radius2 / 2 if self.make_rope2 else radius2

        # generate curves
        curve_obj1 = generate_curve(id=1, points=path_1, radius=radius1, color=color1_value, striped=striped1)
        curve_obj2 = generate_curve(id=2, points=path_2, radius=radius2, color=color2_value, striped=striped2)

        # generate ropes
        if self.make_rope1:
            generate_rope_from_curve(
                curve_obj_name=curve_obj1.name,
                curve_len=np.sum(np.linalg.norm(np.diff(path_1, axis=0), axis=1)),
                radius=radius1 * 2,
                color=color1_value,
                num_chords=2,
            )

        if self.make_rope2:
            generate_rope_from_curve(
                curve_obj_name=curve_obj2.name,
                curve_len=np.sum(np.linalg.norm(np.diff(path_2, axis=0), axis=1)),
                radius=radius2 * 2,
                color=color2_value,
                num_chords=2,
            )

        # get GT points of each curve
        gt_data = self.get_gt_dict([curve_obj1.data, curve_obj2.data], [color1_name, color2_name], [1, 2])
        gt_data["curve_1"]["object_name"] = "cable" if not self.make_rope1 else "rope"
        gt_data["curve_2"]["object_name"] = "cable" if not self.make_rope2 else "rope"

        # Deselect all objects
        bpy.ops.object.select_all(action="DESELECT")

        # Convert to mesh
        for o in bpy.data.objects:
            o.select_set(True)
        bpy.ops.object.convert(keep_original=False)

        return gt_data, good_intersect_points

    def clean_space(self):
        try:
            bpy.data.objects["Cube"].select_set(True)
            bpy.ops.object.delete()
            bpy.data.objects["Camera"].select_set(True)
            bpy.ops.object.delete()
            bpy.data.objects["Light"].select_set(True)
            bpy.ops.object.delete()
        except:
            pass
